@@ -2,37 +2,62 @@ package CGI::Application::Dispatch;
 use strict;
 use warnings;
 use Carp;
+use Exception::Class::TryCatch qw(catch);
 
-our $VERSION = '2.00_04';
-my ($MP, $MP2);
+our $VERSION = '2.00_05';
 our $DEBUG = 0;
 
 # Used for error handling
-my $MODULE_NAME;
 
 # a cache
 our %URL_DISPATCH_CACHE = ();
+our $URL_DISPATCH_CACHE_MAX = 1000;
 
 BEGIN {
-    if( $ENV{MOD_PERL} ) {
-        $MP = 1;
-        $MP2 = exists $ENV{MOD_PERL_API_VERSION} ? $ENV{MOD_PERL_API_VERSION} == 2 : 0;
-        if( $MP2 ) {
-            require Apache2::Const;
-            require Apache2::RequestUtil;
-            require Apache2::RequestRec;
-            require APR::Table;
-        } else {
-            require Apache::Constants;
-        }
+    use constant IS_MODPERL  => exists( $ENV{MOD_PERL} );
+    use constant IS_MODPERL2 => (
+        IS_MODPERL()
+        and exists $ENV{MOD_PERL_API_VERSION} 
+        and $ENV{MOD_PERL_API_VERSION} == 2
+    );
+
+    if (IS_MODPERL2() ) {
+        require Apache2::RequestUtil;
+        require Apache2::RequestRec;
+        require APR::Table;
+        require Apache2::Const;
+        Apache2::Const->import(qw(OK SERVER_ERROR HTTP_BAD_REQUEST NOT_FOUND REDIRECT));
+    } elsif (IS_MODPERL()) {
+        require Apache::Constants;
+        Apache::Constants->import(qw(OK SERVER_ERROR BAD_REQUEST NOT_FOUND REDIRECT));
     }
 }
 
+# these return values have different values used in different ENV
+use Exception::Class (
+    'CGI::Application::Dispatch::Exception',
+    'CGI::Application::Dispatch::ERROR' => {
+        isa         => 'CGI::Application::Dispatch::Exception',
+        alias       => 'throw_error',
+        description => (IS_MODPERL() ? SERVER_ERROR() : 500 ),
+    },
+    'CGI::Application::Dispatch::NOT_FOUND' => {
+        isa         => 'CGI::Application::Dispatch::Exception',
+        alias       => 'throw_not_found',
+        description => ( IS_MODPERL() ? NOT_FOUND() : 404 ),
+    },
+    'CGI::Application::Dispatch::BAD_REQUEST' => {
+        isa         => 'CGI::Application::Dispatch::Exception',
+        alias       => 'throw_bad_request',
+        description => ( IS_MODPERL2() ? HTTP_BAD_REQUEST() : IS_MODPERL ? BAD_REQUEST : 400 ),
+    },
+);
+
 =pod
 
 =head1 NAME
 
-CGI::Application::Dispatch - Dispatch requests to CGI::Application based objects 
+CGI::Application::Dispatch - Dispatch requests to CGI::Application based objects
 
 =head1 SYNOPSIS
 
@@ -56,7 +81,7 @@ Under normal cgi
 
     package MyApp::Dispatch;
     use base 'CGI::Application::Dispatch';
-    
+
     sub args_to_dispatch {
         return {
             prefix  => 'MyApp',
@@ -82,10 +107,10 @@ Under normal cgi
     use MyApp::Dispatch;
     MyApp::Dispatch->dispatch();
 
-=head1 DESCRIPTION 
+=head1 DESCRIPTION
 
-This module provides a way (as a mod_perl handler or running under vanilla CGI) to look at 
-the path (C<< $r->path_info >> or C<$ENV{PATH_INFO}>) of the incoming request, parse 
+This module provides a way (as a mod_perl handler or running under vanilla CGI) to look at
+the path (C<< $r->path_info >> or C<$ENV{PATH_INFO}>) of the incoming request, parse
 off the desired module and it's run mode, create an instance of that module and run it.
 
 It currently supports both generations of mod_perl (1.x and 2.x). Although, for simplicity,
@@ -94,16 +119,16 @@ This may change as mp2 usage increases.
 
 It will translate a URI like this (under mod_perl):
 
-	/app/module_name/run_mode
+    /app/module_name/run_mode
 
 or this (vanilla cgi)
 
-	/app/index.cgi/module_name/run_mode
+    /app/index.cgi/module_name/run_mode
 
 into something that will be functionally similar to this
 
-	my $app = Module::Name->new(..);
-	$app->mode_param(sub {'run_mode'}); #this will set the run mode
+    my $app = Module::Name->new(..);
+    $app->mode_param(sub {'run_mode'}); #this will set the run mode
 
 =head1 METHODS
 
@@ -127,8 +152,8 @@ This method accepts the following name value pairs:
 
 =item default
 
-This option will set a default value if there is no C<PATH_INFO>. It will be parsed
-to obtain the module name and run mode.
+Specify a value to use in place of C<PATH_INFO> if C<PATH_INFO> is not provided.
+This could be the case if the default page is selected (eg: "/" ).
 
 =item prefix
 
@@ -147,14 +172,24 @@ This is a hash of arguments that are passed into the C<new()> constructor of the
 
 =item table
 
-In most cases, simply using Dispatch with the C<default> and C<prefix> is enough 
-to simplify your application and your URLs, but there are many cases where you want 
+In most cases, simply using Dispatch with the C<default> and C<prefix> is enough
+to simplify your application and your URLs, but there are many cases where you want
 more power. Enter the dispatch table. Since this table can be slightly complicated,
 a whole section exists on it's use. Please see the L<DISPATCH TABLE> section.
 
 =item debug
 
-Send debugging output for this module to STDERR. 
+Set to a true value to send debugging output for this module to STDERR. Off by default. 
+
+=item not_found
+
+This is a URI to which the application will be redirected if the incoming request's
+C<PATH_INFO> does not match anything in the L<DISPATCH TABLE>. This will also be used
+if there was a match, but the module could not be loaded, or the run mode did not
+exist.
+
+If this value is not present, then Dispatch will return a NOT FOUND error either to the 
+browser (under CGI) or to Apache (under mod_perl).
 
 =back
 
@@ -168,30 +203,25 @@ sub dispatch {
     $DEBUG = 1 if $args{debug};
 
     # Immediatey, try a cached result.
-    if (
-        defined $ENV{REQUEST_URI}
-        &&
-        $ENV{REQUEST_URI}
-        &&
-        ( my $final_args = $self->_get_cache($ENV{REQUEST_URI} ) )
-    ) {
+    if ( my $final_args = $self->_url_cache() ) {
         if( $DEBUG ) {
             require Data::Dumper;
-            warn "[Dispatch] - Found cached version of URL '$ENV{REQUEST_URI}'. Using the following args: "
-                . Data::Dumper::Dumper($final_args);
+            warn "[Dispatch] - Found cached version of URL '$ENV{REQUEST_URI}'. "
+                . "Using the following args: " . Data::Dumper::Dumper($final_args);
         }
         return $self->_run_app(@$final_args);
     }
 
     # check for extra args (for backwards compatibility)
     foreach (keys %args) {
-        next if( 
-            $_ eq 'prefix' || 
-            $_ eq 'default' || 
-            $_ eq 'debug' || 
-            $_ eq 'rm' || 
-            $_ eq 'args_to_new' || 
-            $_ eq 'table' 
+        next if(
+            $_ eq 'prefix'      or
+            $_ eq 'default'     or
+            $_ eq 'debug'       or
+            $_ eq 'rm'          or
+            $_ eq 'args_to_new' or
+            $_ eq 'table'       or
+            $_ eq 'not_found'
         );
         carp "Passing extra args ('$_') to dispatch() is deprecated! Please use 'args_to_new'";
         $args{args_to_new}->{$_} = delete $args{$_};
@@ -202,162 +232,276 @@ sub dispatch {
     my $path_info = $ENV{PATH_INFO};
     # use the 'default' if we need to
     $path_info = $args{default} || '' if( !$path_info || $path_info eq '/' );
-    # make sure they all start with a '/', to correspond with the RE we'll make
+    # make sure they all start and end with a '/', to correspond with the RE we'll make
     $path_info = "/$path_info" unless( index($path_info, '/') == 0 );
     $path_info = "$path_info/" unless( substr($path_info, -1) eq '/');
 
+    my ( $module, $rm, $local_prefix, $local_args_to_new );
+
+    # take args from path
+    my $named_args = $self->_parse_path( $path_info, $args{table} )
+		or throw_not_found("Could not parse PATH_INFO");
+
+    if ($DEBUG) {
+        require Data::Dumper;
+        warn "[Dispatch] Named args from match: "
+          . Data::Dumper::Dumper($named_args) . "\n";
+    }
+
+    if (exists($named_args->{PARAMS}) || exists($named_args->{TMPL_PATH})) {
+        carp "PARAMS and TMPL_PATH are not allowed here. Did you mean to use args_to_new?";
+        throw_error("PARAMS and TMPL_PATH not allowed");
+    }
+
+    # eval and catch any exceptions that might be thrown
+    my ($output, @final_dispatch_args);
+    eval {
+        ( $module, $local_prefix, $rm, $local_args_to_new ) =
+          delete @{$named_args}{qw(app prefix rm args_to_new)};
+
+        $module or throw_error("App not defined");
+        $module = $self->translate_module_name($module);
+
+        $local_prefix ||=  $args{prefix};
+        $module = $local_prefix . '::' . $module if( $local_prefix );
+
+        $local_args_to_new ||= $args{args_to_new};
+
+        # add the rest of the named_args to PARAMS
+        @{ $local_args_to_new->{PARAMS} }{ keys %$named_args } = values %$named_args;
+
+        # load and run the module
+        @final_dispatch_args = ($module, $rm, $local_args_to_new);
+        $self->require_module($module);
+        $output = $self->_run_app($module,$rm,$local_args_to_new);
+    };
+    my $e;
+    catch( $e, [ 'CGI::Application::Dispatch::Exception' ])
+        and return $self->http_error($e, $args{not_found});
+
+    # Cache this URL - dispatch map for later use.
+    $self->_url_cache(\@final_dispatch_args);
+
+    return $output;
+}
+
+
+# stolen from http_protocol.c in Apache sources
+# we don't actually use anything other than 200, 307, 400, 404 and 500
+my %status_lines = (
+#    100 => 'Continue',
+#    101 => 'Switching Protocols',
+#    102 => 'Processing',
+    200 => 'OK',
+#    201 => 'Created',
+#    202 => 'Accepted',
+#    203 => 'Non-Authoritative Information',
+#    204 => 'No Content',
+#    205 => 'Reset Content',
+#    206 => 'Partial Content',
+#    207 => 'Multi-Status',
+#    300 => 'Multiple Choices',
+#    301 => 'Moved Permanently',
+#    302 => 'Found',
+#    303 => 'See Other',
+#    304 => 'Not Modified',
+#    305 => 'Use Proxy',
+    307 => 'Temporary Redirect',
+    400 => 'Bad Request',
+#    401 => 'Authorization Required',
+#    402 => 'Payment Required',
+#    403 => 'Forbidden',
+    404 => 'Not Found',
+#    405 => 'Method Not Allowed',
+#    406 => 'Not Acceptable',
+#    407 => 'Proxy Authentication Required',
+#    408 => 'Request Time-out',
+#    409 => 'Conflict',
+#    410 => 'Gone',
+#    411 => 'Length Required',
+#    412 => 'Precondition Failed',
+#    413 => 'Request Entity Too Large',
+#    414 => 'Request-URI Too Large',
+#    415 => 'Unsupported Media Type',
+#    416 => 'Requested Range Not Satisfiable',
+#    417 => 'Expectation Failed',
+#    422 => 'Unprocessable Entity',
+#    423 => 'Locked',
+#    424 => 'Failed Dependency',
+    500 => 'Internal Server Error',
+#    501 => 'Method Not Implemented',
+#    502 => 'Bad Gateway',
+#    503 => 'Service Temporarily Unavailable',
+#    504 => 'Gateway Time-out',
+#    505 => 'HTTP Version Not Supported',
+#    506 => 'Variant Also Negotiates',
+#    507 => 'Insufficient Storage',
+#    510 => 'Not Extended',
+);
+
+sub http_error {
+    my ($self, $e, $url) = @_;
+
+    warn "[Dispatch] ERROR: '" . $e->error 
+        . ( $ENV{REQUEST_URI} ? "' for request '$ENV{REQUEST_URI}'" : "' ") . "\n";
+    my $redirect;
+    if( $e->isa('CGI::Application::Dispatch::NOT_FOUND') && $url ) {
+        $redirect = 1;
+        warn "[Dispatch] Redirection for NOT FOUND to $url" if( $DEBUG );
+    }
+    my $errno = $e->description;
+
+    # if we're under mod_perl
+    if (IS_MODPERL) {
+        my $r = IS_MODPERL2 ? Apache2::RequestUtil->request : Apache->request;
+        # if we just want to redirect
+        if( $redirect ) {
+            $r->status( $errno );
+            $r->headers_out->{'Location'} = $url; 
+        } else {
+            # next line is for Apache::Registry
+            $r->status($errno);
+        }
+        return '';
+    # else print the HTTP stuff ourselves
+    } else {
+        $errno = 500 if( !exists $status_lines{$errno} );
+        my $output = "HTTP/1.0 Status: $errno $status_lines{$errno}\n";
+        $output .= "Location: $url\n" if( $redirect );
+        $output .= "Content-type: text/html\n\n";
+
+        # TODO: possibly provide more feedback in a way that is XSS safe.
+        # (I'm not sure that passing through the raw ENV variable directly is safe.)
+        # <P>We tried: $ENV{REQUEST_URI}</P></BODY></HTML>";
+        $output .= qq(
+            <!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
+            <HTML><HEAD>
+            <TITLE>$errno $status_lines{$errno}</TITLE>
+            </HEAD><BODY>
+            <H1>ERROR: $status_lines{$errno}</H1>
+            <P>Please contact the server administrator )
+            . ( $ENV{SERVER_ADMIN} ? "($ENV{SERVER_ADMIN})" : '') 
+            . qq( and inform them
+            of the time the error occurred, and anything you might have done that may have
+            caused the error.</P>
+            <HR>)
+            . ( $ENV{SERVER_SIGNATURE} || '' )
+            . qq(</BODY></HTML>);
+        return $output;
+    }
+}
+
+
+
+# protected method - designed to be used by sub classes, not by end users
+sub _parse_path {
+    my ( $self, $path, $table ) = @_;
+
     # get the module name from the table
-    my $table = $args{table} or croak "Must at least have a default 'table'!";
-    my $table_index;
-    my ($module, $rm, $local_args_to_new);
-    for(my $i = 0; $i < scalar (@$table); $i+=2) {
+    return unless defined($path);
+
+    unless (ref($table) eq 'ARRAY' ) {
+        warn "[Dispatch] Invalid or no dispatch table!\n";
+        return;
+    }
+
+    for ( my $i = 0 ; $i < scalar(@$table) ; $i += 2 ) {
+
         # translate the rule into a regular expression, but remember where the named args are
         my $rule = $table->[$i];
+
         # make sure they start and end with a '/' to match how PATH_INFO is formatted
-        $rule = "/$rule" unless( index($rule, '/') == 0 );  
-        $rule = "$rule/" unless( substr($rule, -1) eq '/');
-        my ($regex, @names);
-        # '/:foo' will become '/([^\/]*)' 
+        $rule = "/$rule" unless ( index( $rule, '/' ) == 0 );
+        $rule = "$rule/" if ( substr( $rule, -1 ) ne '/' );
+
+        my @names = ();
+
+        # '/:foo' will become '/([^\/]*)'
         # and
         # '/:bar?' will become '/?([^\/]*)?'
         # and then remember which position it matches
-        { 
-            # remove warning about $4 being used in the result when it doesn't always match
-            no warnings; 
-            # TODO: document what $1 - $4 mean.
-            while( $rule =~ s{
-                    (^|/)                 # beginning or a /
-                    (:([^/\?]+)(\?)?)     # stuff in between 
-                }
-                {$1$4([^/]*)$4}x ) {
-                push(@names, $3); # it's the 3rd grouping from the match above
-            }
-        }
-        # make sure we only match this rule
-        $rule = '^' . $rule . '$';
 
-        if( $DEBUG ) {
-            warn "[Dispatch] Trying to match '$path_info' against rule '$table->[$i]' "
-                . "using regex '$rule'\n";
-        }
+        $rule =~ s{
+            (^|/)                 # beginning or a /
+            (:([^/\?]+)(\?)?)     # stuff in between
+        }{
+            push(@names, $3);
+            $1 . ($4 ? '?([^/]*)?' : '([^/]*)')
+        }gxe;
+
+        warn "[Dispatch] Trying to match '${path}' against rule '$table->[$i]' using regex '${rule}'\n"
+			if $DEBUG;
 
         # if we found a match, then run with it
-        if( $path_info =~ /$rule/ ) {
-            warn "[Dispatch] Matched!\n" if( $DEBUG );
+        if ( my @values = ( $path =~ m#^$rule$# ) ) {
 
-            my $named_args = $table->[$i+1];
-            # add the extra named_args from the match
-            for(my $j = 0; $j<=$#names; $j++) {
-                no strict 'refs';
-                $named_args->{$names[$j]} = ${$j +1};
-            }
+            warn "[Dispatch] Matched!\n" if $DEBUG;
 
-            if( $DEBUG ) {
-                require Data::Dumper;
-                warn "[Dispatch] Named args from match: " . Data::Dumper::Dumper($named_args) . "\n";
-            }
-            my $module_name = $named_args->{app} || croak "No 'app' contained in the match!";
-            $module = $self->translate_module_name($module_name);
-            # now add the prefix
-            my $local_prefix = $named_args->{prefix} || $args{prefix};
+            my %named_args = %{ $table->[ ++$i ] };
+            @named_args{@names} = @values  if @names;
 
-            $module = $local_prefix . '::' . $module if( $local_prefix );
-
-            if (defined $named_args->{'args_to_new'}) {
-                $local_args_to_new = $named_args->{'args_to_new'};
-            }
-            else {
-                $local_args_to_new = $args{args_to_new};
-            }
-
-            # add the rest of the named_args to PARAMS
-            foreach my $named (keys %$named_args) {
-                if ($named =~ m/^PARAMS|TMPL_PATH/) {
-                    croak "PARAMS and TMPL_PATH are not allowed here. Did you mean to use args_to_new?";
-                }
-                next if ($named =~ m/^rm|app$/);
-                $local_args_to_new->{PARAMS}->{$named} = $named_args->{$named};
-            }
-
-            # Use local args to new, or default to the global
-
-            # remember the rm if we have one
-            $rm = $named_args->{rm};
-
-            last;
+            return \%named_args;
         }
-    };
-
-    if (defined $module and length $module) {
-        my @final_dispatch_args = ($module,$rm,$local_args_to_new);
-
-        # Cache this URL - dispatch map for later use.
-        $self->_set_cache($ENV{REQUEST_URI} => \@final_dispatch_args)
-            if( $ENV{REQUEST_URI} );
-
-        return $self->_run_app(@final_dispatch_args);
-
-    }
-    # Return 404 here if module isn't found 
-    # but not for mod_perl, which handles 404 stuff itself. 
-    elsif  (not $ENV{MOD_PERL} ) {
-        warn "[Dispatch] 404 for $ENV{REQUEST_URI}" if $DEBUG;
-        print "Status: 404 Not Found\n";
-        print "Content-type: text/html\n\n";
-        print q{<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
-                <HTML><HEAD><TITLE>File Not Found</TITLE></HEAD><BODY>
-                <H1>File Not Found.</H1></BODY></HTML>};
-        # TODO: possibly provide more feedback in a way that is XSS safe.
-        # (I'm not sure that passing through the raw ENV variable directly is safe.) 
-        # <P>We tried: $ENV{REQUEST_URI}</P></BODY></HTML>";
     }
+
+    return;
 }
 
 # protected method - designed to be used by sub classes, not by end users
-sub _set_cache {
-    my ($class, $key, $value) = @_;
-    $URL_DISPATCH_CACHE{$key} = $value;
-}
+sub _url_cache {
+    my ( $self, $value ) = @_;
+    my $key;
+
+    # may be PATH_TRANSLATED only to use here?
+    foreach my $env qw(SERVER_NAME SCRIPT_NAME PATH_INFO) {
+        $key .= ":" . ($ENV{$env} || '');
+    }
+
+    if ( defined($value) ) {
+        # shrink cache, if it's too large
+        delete(
+            @URL_DISPATCH_CACHE{
+                ( keys %URL_DISPATCH_CACHE )
+                  [ 0 .. int( $URL_DISPATCH_CACHE_MAX / 3 ) ]
+              }
+          )
+          if ( keys %URL_DISPATCH_CACHE > $URL_DISPATCH_CACHE_MAX );
+
+        $URL_DISPATCH_CACHE{$key} = $value;
+    }
 
-# protected method - designed to be used by sub classes, not by end users
-sub _get_cache {
-    my ($class, $key) = @_;
     return $URL_DISPATCH_CACHE{$key};
 }
 
 sub _run_app {
-    my ($self,$module,$rm,$args) = @_;
+    my ( $self, $module, $rm, $args ) = @_;
 
-    if( $DEBUG ) {
+    if ($DEBUG) {
         require Data::Dumper;
-        warn "[Dispatch] Final args to pass to new(): " . Data::Dumper::Dumper($args) . "\n";
-    }
-    
-    # now create and run the application object
-    $MODULE_NAME = $module;
-    warn "[Dispatch] creating instance of $module\n" if( $DEBUG );
-    $self->require_module($module);
-    my $app;
-    eval {
-        if( defined $args && %{$args} ) {
-            $app = $module->new($args);
-        } else {
-            $app = $module->new();
-        }
-    };
-    if ($@) {
-        my $msg = "Unable to run '$module': $@";
-        warn "[Dispatch] $msg" if( $DEBUG );
-        croak $msg;
+        warn "[Dispatch] Final args to pass to new(): "
+          . Data::Dumper::Dumper($args) . "\n";
     }
 
-    $app->mode_param(sub { return $rm }) if( $rm );
-    $app->run();
+    # now create and run then application object
+    warn "[Dispatch] creating instance of $module\n" if( $DEBUG );
+
+    my $output;
+    eval {
+        my $app = ref($args) eq 'HASH' ? $module->new($args) : $module->new();
+        $app->mode_param( sub { return $rm } ) if ($rm);
+        $output = $app->run();
+    };
+
+    if ($@) {
+        throw_not_found('RM not found') if ( $@ =~ /No such run mode/ );
+    }
+
+    return $output;
 }
 
 =head2 handler()
 
-This method is used so that this module can be run as a mod_perl handler. 
+This method is used so that this module can be run as a mod_perl handler.
 When it creates the application module it passes the $r argument into the PARAMS
 hash of new()
 
@@ -381,13 +525,13 @@ as 'r'.
     # inside your app
     my $request = $self->param('r');
 
-If you need more customization than can be accomplished with just L<prefix> 
+If you need more customization than can be accomplished with just L<prefix>
 and L<default>, then it would be best to just subclass CGI::Application::Dispatch
 and override L<dispatch_args> since this method uses L<dispatch> to do the heavy lifting.
 
     package MyApp::Dispatch;
     use base 'CGI::Application::Dispatch';
-    
+
     sub dispatch_args {
         return {
             prefix  => 'MyApp',
@@ -418,9 +562,9 @@ And then in your httpd.conf
 
 sub handler : method {
     my ($self, $r) = @_;
-    
+
     # set the PATH_INFO
-    $ENV{PATH_INFO} ||= $r->path_info();
+    $ENV{PATH_INFO} = $r->path_info();
 
     # setup our args to dispatch()
     my %args;
@@ -437,33 +581,12 @@ sub handler : method {
     $DEBUG = 1 if( $config_args->{CGIAPP_DISPATCH_DEBUG} );
     if( $DEBUG ) {
         require Data::Dumper;
-        warn "[Dispatch] Calling dispatch() with the following arguments: " 
-            . Data::Dumper::Dumper(\%args) . "\n";
-    }
-    eval { $self->dispatch(%args) };
-
-    #if we had an error
-    if ($@) {
-        #let's check to see if that module could not be found
-        my $module_path = $MODULE_NAME;
-        $module_path =~ s/::/\//g;
-
-        if ( $@ =~ /Can't locate $module_path.pm/ ) {
-            warn "CGI::Application::Dispatch - Couldn't find module $MODULE_NAME";
-            return $MP2 ? Apache2::Const::NOT_FOUND() : Apache::Constants::NOT_FOUND();
-        # else if the run mode doesn't exist
-        } elsif( $@ =~ /No such run mode/ ) {
-            warn "CGI::Application::Dispatch - Couldn't find runmode";
-            return $MP2 ? Apache2::Const::NOT_FOUND() : Apache::Constants::NOT_FOUND();
-        }
-        #else there was some other error
-        else {
-            warn "CGI::Application::Dispatch - ERROR $@";
-            return $MP2 ? Apache2::Const::SERVER_ERROR() : Apache::Constants::SERVER_ERROR();
-        }
+        warn "[Dispatch] Calling dispatch() with the following arguments: " . Data::Dumper::Dumper(\%args) . "\n";
     }
 
-    return $MP2 ? Apache2::Const::OK() : Apache::Constants::OK();
+    $self->dispatch(%args);
+
+    return $r->status || OK();
 }
 
 =head2 dispatch_args
@@ -501,11 +624,11 @@ sub dispatch_args {
     };
 }
 
-=head2 translate_module_name 
+=head2 translate_module_name
 
 This method is used to control how the module name is translated from
 the matching section of the C<PATH_INFO> (see L<"PATH_INFO Parsing">.
-The main reason that this method exists is so that it can be overridden if it doesn't do 
+The main reason that this method exists is so that it can be overridden if it doesn't do
 exactly what you want.
 
 The following transformations are performed on the input:
@@ -554,19 +677,27 @@ any problems requiring the named module, then we will C<croak>.
 
 sub require_module {
     my ($self, $module) = @_;
-    if( $module ) {
-        #untaint the module name
-        ($module) = ($module =~ /^([A-Za-z][A-Za-z0-9_\-\:\']+)$/);   
-        croak "Invalid characters used in module name" unless ($module);
-        eval "require $module";
-    
-        if( $@ ) {
-            warn "[Dispatch] Unable to load module '$module': $@" if( $DEBUG );
-            croak $@;
-        }
-        return $module;
+
+    $module or die "STATUS: " . DISPATCH_NOT_FOUND();
+
+    #untaint the module name
+    ($module) = ($module =~ /^([A-Za-z][A-Za-z0-9_\-\:\']+)$/);
+
+    unless ($module) {
+        throw_bad_request("Invalid characters in module name");
+    }
+
+    warn "[Dispatch] loading module $module\n" if ( $DEBUG );
+    eval "require $module";
+    return unless $@;
+
+    my $module_path = $module;
+    $module_path =~ s/::/\//g;
+
+    if ( $@ =~ /Can't locate $module_path.pm/ ) {
+        throw_not_found("Can't find module $module");
     } else {
-        return;
+        throw_error("Unable to load module '$module': $@");
     }
 }
 
@@ -588,17 +719,17 @@ Sometimes it's easiest to explain with an example, so here you go:
     table       => [
         ''                         => { app => 'Blog', rm => 'recent'}
         'posts/:category'          => { app => 'Blog', rm => 'posts' },
-        ':app/:rm/:id'             => { app => 'Blog' },  
-        'date/:year/:month?/:day?' => { 
-            app         => 'Blog', 
-            rm          => 'by_date', 
+        ':app/:rm/:id'             => { app => 'Blog' },
+        'date/:year/:month?/:day?' => {
+            app         => 'Blog',
+            rm          => 'by_date',
             args_to_new => { TMPL_PATH = "events/" },
         },
     ]
   );
 
 So first, this call to L<dispatch> set's the L<prefix> and passes a C<TMPL_PATH>
-into L<args_to_new>. Next it sets the L<table>. 
+into L<args_to_new>. Next it sets the L<table>.
 
 
 =head2 VOCABULARY
@@ -613,7 +744,7 @@ is called a C<token>.
 
 =head2 URL MATCHING
 
-When a URL comes in, Dispatch tries to match it against each rule in the table in 
+When a URL comes in, Dispatch tries to match it against each rule in the table in
 the order in which the rules are given. The first one to match wins.
 
 A rule consists of backslashes and tokens. A token can one of the following types:
@@ -667,8 +798,8 @@ optional. If the rest of the URL matches the rest of the rule, then it doesn't m
 it contains this token or not. It's best to only include optional-variable tokens at the end
 of your rule. In the rule
 
-    'date/:year/:month?/:day?' 
-    
+    'date/:year/:month?/:day?'
+
 C<:month?> and C<:day?> are optional-variable tokens.
 
 Just like with L<variable> tokens, optional-variable tokens' values can also be retrieved by
@@ -696,25 +827,25 @@ For instance, if you want to override L<prefix> for a specific rule, then you ca
 =head1 PATH_INFO Parsing
 
 This section will describe how the application module and run mode are determined from
-the C<PATH_INFO> if no L<DISPATCH TABLE> is present, and what options you have to 
+the C<PATH_INFO> if no L<DISPATCH TABLE> is present, and what options you have to
 customize the process.
 
 =head2 Getting the module name
 
-To get the name of the application module the C<PATH_INFO> is split on backslahes (C</>). 
-The second element of the returned list is used to create the application module. So if we 
+To get the name of the application module the C<PATH_INFO> is split on backslahes (C</>).
+The second element of the returned list is used to create the application module. So if we
 have a path info of
 
     /module_name/mode1
 
 then the string 'module_name' is used. This is passed through the L<translate_module_name>
-method. Then if there is a C<prefix> (and there should always be a L<prefix>) it is added 
-to the beginning of this new module name with a double colon C<::> separating the two. 
+method. Then if there is a C<prefix> (and there should always be a L<prefix>) it is added
+to the beginning of this new module name with a double colon C<::> separating the two.
 
-If you don't like the exact way that this is done, don't fret you do have a couple of options. 
+If you don't like the exact way that this is done, don't fret you do have a couple of options.
 First, you can specify a L<DISPATCH TABLE> which is much more powerfule and flexible (in fact
 this default behavior is actually implemented internally with a dispatch table).
-Or if you want something a little simpler, you can simply subclass and extend the 
+Or if you want something a little simpler, you can simply subclass and extend the
 L<translate_module_name> method.
 
 =head2 Getting the run mode
@@ -735,7 +866,7 @@ Then the string 'mode2' is used as the run mode.
 
 CGI query strings are unaffected by the use of C<PATH_INFO> to obtain the module name and run mode.
 This means that any other modules you use to get access to you query argument (ie, L<CGI>,
-L<Apache::Request>) should not be affected. But, since the run mode may be determined by 
+L<Apache::Request>) should not be affected. But, since the run mode may be determined by
 CGI::Application::Dispatch having a query argument named 'rm' will be ignored by your application
 module.
 
@@ -748,13 +879,38 @@ With a dispatch script, you can fairly clean URLS like this:
  /cgi-bin/dispatch.cgi/module_name/run_mode
 
 However, including "/cgi-bin/dispatch.cgi" in ever URL doesn't add any value to the URL,
-so it's nice to remove it. This is easily done if you are using the Apache web server with 
+so it's nice to remove it. This is easily done if you are using the Apache web server with
 C<mod_rewrite> available. Adding the following to a C<.htaccess> file would allow you to
 simply use:
 
  /module_name/run_mode
 
-.htaccess file example
+If you have problems with mod_rewrite, turn on debugging to see exactly what's happening:
+
+ RewriteLog /home/project/logs/alpha-rewrite.log
+ RewriteLogLevel 9
+
+=head2 mod_rewrite related code in the dispatch script.
+
+This seemed necessary to put in the dispatch script to make mod_rewrite happy.
+Perhaps it's specific to using C<RewriteBase>. Refactorings welcome!
+
+  # mod_rewrite alters the PATH_INFO by turning it into a file system path,
+  # so we repair it.
+  use FindBin qw/$Bin/;
+  use Cwd 'abs_path';
+  chdir $Bin;
+  # Location of the www root in relation to the dispatcher. (Adjust as needed!)
+  my $www = abs_path("../");
+  # fix /home versus /usr/home difference
+  $www =~ s{/usr}{};
+
+  # the final removal of the file system stuff mod_rewrite added.
+  # PATH_INFO won't defined if "/" is requested.
+  $ENV{PATH_INFO} =~ s/^$www//g if defined $ENV{PATH_INFO};
+
+
+=head2 Simple Apache Example
 
   RewriteEngine On
 
@@ -762,13 +918,59 @@ simply use:
   # specific directory.
   RewriteBase /
 
-  # If an actual file or directory is requested, serve directly    
+  # If an actual file or directory is requested, serve directly
   RewriteCond %{REQUEST_FILENAME} !-f
   RewriteCond %{REQUEST_FILENAME} !-d
 
   # Otherwise, pass everything through to the dispatcher
   RewriteRule ^(.*)$ /cgi-bin/dispatch.cgi$1 [L,QSA]
 
+=head2 More complex rewrite: dispatching "/" and multiple developers
+
+Here is a more complex example that dispatches "/", which would otherwise
+be treated as a directory, and also supports multiple developer directories,
+so C</~mark> has its own seperate dispatching system beneath it.
+
+Note that order matters here! The Location block for "/" needs to come before the
+user blocks.
+
+  <Location />
+    RewriteEngine On
+    RewriteBase /
+
+    # Run "/" through the dispatcher
+    RewriteRule ^home/project/www/$ /cgi-bin/dispatch.cgi [L,QSA]
+
+    # Don't apply this rule to the users sub directories.
+    RewriteCond %{REQUEST_URI} !^/~.*$
+    # If an actual file or directory is requested, serve directly
+    RewriteCond %{REQUEST_FILENAME} !-f
+    RewriteCond %{REQUEST_FILENAME} !-d
+    # Otherwise, pass everything through to the dispatcher
+    RewriteRule ^(.*)$ /cgi-bin/dispatch.cgi/$1 [L,QSA]
+  </Location>
+
+  <Location /~mark>
+    RewriteEngine On
+    RewriteBase /~mark
+
+    # Run "/" through the dispatcher
+    RewriteRule ^/home/mark/www/$ /~mark/cgi-bin/dispatch.cgi [L,QSA]
+
+     # Otherwise, if an actual file or directory is requested, serve directly
+    RewriteCond %{REQUEST_FILENAME} !-f
+    RewriteCond %{REQUEST_FILENAME} !-d
+
+     # Otherwise, pass everything through to the dispatcher
+    RewriteRule ^(.*)$ /~mark/cgi-bin/dispatch.cgi$1 [L,QSA]
+
+    # These examples may also be helpful, but are unrelated to dispatching.
+    SetEnv DEVMODE mark
+    SetEnv PERL5LIB /home/mark/perllib:/home/mark/config
+    ErrorDocument 404 /~mark/errdocs/404.html
+    ErrorDocument 500 /~mark/errdocs/500.html
+  </Location>
+
 =head1 SUBCLASSING
 
 While Dispatch tries to be flexible, it won't be able to do everything that people want. Hopefully
@@ -830,7 +1032,7 @@ L<http://www.cgi-app.org/>
 
 =item * Mark Stosberg <mark@summersault.com>
 
-=item * Viacheslav Sheveliov <vsheveliov@mail.ru>
+=item * Viacheslav Sheveliov <slavash@aha.ru>
 
 =back
 
@@ -840,7 +1042,7 @@ Since C::A::Dispatch will dynamically choose which modules to use as the content
 it may give someone the ability to execute random modules on your system if those modules can
 be found in you path. Of course those modules would have to behave like L<CGI::Application> based
 modules, but that still opens up the door more than most want. This should only be a problem
-if you don't use a L<prefix>. By using this option you are only allowing Dispatch to pick from 
+if you don't use a L<prefix>. By using this option you are only allowing Dispatch to pick from
 a namespace of modules to run.
 
 =head1 SEE ALSO
